# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

import os.path
from fiona import supported_drivers

suffix_map = {
    'gdb': 'FileGDB',
    'shp': 'ESRI Shapefile',
    'gtm': 'GPSTrackMaker',
    'dgn': 'DGN',
    'json': 'GeoJSON',
    'geojson': 'GeoJSON',
    'gpkg': 'GPKG',
    'mapinfo': 'MapInfo File',
    'gpx': 'GPX',
    'dxf': 'DXF',
    'bna': 'BNA',
    'gmt': 'GMT',
}


def from_path(path):
    _, suffix = os.path.splitext(os.path.basename(path))
    return from_suffix(suffix)


def from_suffix(suffix):
    '''Attempt to return the name of the appropriate GDAL driver, given a suffix.'''

    if suffix.startswith('.'):
        suffix = suffix[1:]

    if suffix in suffix_map and suffix_map[suffix] in supported_drivers:
        return suffix_map[suffix]

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from . import test_coords
from . import test_feature
from . import test_layer
from . import test_measure
from . import test_geometry
from . import test_round
from . import test_scale



# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

try:
    from shapely.geometry import shape
except ImportError:
    pass


def field_contains_test(field_values):
    '''
    Return a test function that checks if the properties of a feature match the possible values in field_values.

    >>> test = field_contains_test({'a': (1, 2, 3)})
    >>> test({'properties': {'a': 1}})
    True
    >>> test({'properties': {'a': 4}})
    False
    '''
    def test(feature):
        for k, v in field_values.items():
            if feature['properties'].get(k) not in v:
                return False
        return True

    return test


def togeojson(typ, coordinates, properties=None):
    '''Return a GeoJSON-ready object given a properties dict, a type and coordinates.'''
    properties = properties or {}

    return {
        'type': 'Feature',
        'properties': properties,
        'geometry': {
            'type': typ,
            'coordinates': coordinates,
        }
    }


def shapify(feature):
    '''Applies shapely.geometry.shape to the geometry part of a feature
    and returns a new feature object with properties intact'''
    try:
        return {
            'properties': feature.get('properties'),
            'geometry': shape(feature['geometry'])
        }

    except NameError:
        raise NotImplementedError("shapify requires shapely")


def length(feature):
    '''Returns shapely length'''
    try:
        geom = shape(feature['geometry'])
        return geom.length

    except NameError:
        raise NotImplementedError("length requires shapely")


def compound(feature):
    '''Returns True if feature has more than one part:
    * A polygon with holes
    * A MultiPolygon or MultiLineString
    '''
    if 'Multi' in feature['geometry']['type']:
        return True

    if (feature['geometry']['type'] == 'Polygon' and
            len(feature['geometry']['coordinates']) > 1):
        return True

    return False



# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

from .round import round_ring as roundring


def max_x(coords):
    return max((c[0] for c in coords))


def max_y(coords):
    return max((c[1] for c in coords))


def min_x(coords):
    return min((c[0] for c in coords))


def min_y(coords):
    return min((c[1] for c in coords))


def segmentize(ring):
    for i, point in enumerate(ring[:-1]):
        yield point, ring[i + 1]


def bounds(ring):
    '''Return minimum bounding rectangle for a ring'''
    return min_x(ring), min_y(ring), max_x(ring), max_y(ring)


def roundpolyring(polyring, precision=None):
    return [roundring(ring, precision) for ring in polyring]


def centerbounds(bounds):
    '''Returns the center of a bounding box.'''
    return (bounds[2] - bounds[0]) / 2, (bounds[3] - bounds[1]) / 2


def cornerbounds(bounds):
    '''Returns the four corners of a bounding box'''
    return (bounds[0], bounds[1]), (bounds[0], bounds[3]), (bounds[2], bounds[3]), (bounds[2], bounds[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from unittest import TestCase as PythonTestCase
import unittest.main
from fionautil import coords

class TestCoords(PythonTestCase):

    def setUp(self):
        self.coords = [(1, 1), (-2, 2), (3, 14), (10, -1)]

    def testMinX(self):
        assert coords.min_x(self.coords) == -2

    def testMinY(self):
        assert coords.min_y(self.coords) == -1

    def testMaxX(self):
        assert coords.max_x(self.coords) == 10

    def testMaxY(self):
        assert coords.max_y(self.coords) == 14

    def test_segmentize(self):
        segments = coords.segmentize(self.coords)

        self.assertEqual(next(segments), ((1, 1), (-2, 2)))
        self.assertEqual(next(segments), ((-2, 2), (3, 14)))

    def test_bounds(self):
        assert coords.bounds(self.coords) == (-2, -1, 10, 14)

    def test_cornerbounds(self):
        b = (0, 0, 6, 8)
        self.assertSequenceEqual(coords.cornerbounds(b), [(0, 0), (0, 8), (6, 8), (6, 0)])

if __name__ == '__main__':
    unittest.main()



# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

import math

try:
    import pyproj
    WGS84 = pyproj.Geod(ellps='WGS84')
except ImportError:
    pass

def distance(x0, y0, x1, y1, longlat=True):
    '''distance (in m) between two pairs of points'''
    if longlat:
        try:
            _, _, d = WGS84.inv(x0, y0, x1, y1)
        except NameError:
            raise NotImplementedError("Distance of long/lat coordinates requires pyproj.")
    else:
        d = math.sqrt((x0 - x1) ** 2 + (y0 - y1) ** 2)
    return d


def distance_points(p0, p1, longlat=True):
    '''distance (in m) between two points'''
    return distance(p0[0], p0[1], p1[0], p1[1], longlat)


def _projected_azimuth(x0, y0, x1, y1):
    '''The angle of a line between two points on a cartesian plane. Always clockwise and in degrees'''
    if y0 == y1:
        if x0 == x1:
            az = 0.
        elif x1 > x0:
            az = math.pi / 2.
        else:
            az = 3 * math.pi / 2.
    else:
        az = math.atan2((x1 - x0), (y1 - y0))

    return math.degrees(az)


def azimuth(x0, y0, x1, y1, radians=False, clockwise=None, longlat=True):
    '''
    Measure the azimuth between 2 geographic points

    Args:
        x0 (float): first x coordinate.
        y0 (float): first y coordinate.
        x1 (float): second x coordinate.
        y1 (float): second y coordinate.
        radians (bool): Return in radians.
        clockwise (bool): Return with clockwise coordinates.
        longlat (bool): Input is in longitude-latitude, rather than projected.

    Returns:
        float
    '''
    if longlat:
        # this is always in angles
        try:
            az, _, _ = WGS84.inv(x0, y0, x1, y1)
        except NameError:
            raise NotImplementedError("Distance of long/lat coordinates requires pyproj.")

    else:
        az = _projected_azimuth(x0, y0, x1, y1)

    if radians:
        az = math.radians(az)

    return az * (1 if clockwise else -1)


def azimuth_points(p0, p1, radians=False, clockwise=None, longlat=True):
    return azimuth(p0[0], p0[1], p1[0], p1[1], radians, clockwise, longlat)


def signed_area(coords):
    """Return the signed area enclosed by a ring using the linear time
    algorithm at http://www.cgafaq.info/wiki/Polygon_Area. A value >= 0
    indicates a counter-clockwise oriented ring."""

    try:
        xs, ys = tuple(map(list, zip(*coords)))
    except ValueError:
        # Attempt to handle a z-dimension
        xs, ys, _ = tuple(map(list, zip(*coords)))

    xs.append(xs[1])
    ys.append(ys[1])
    return sum(xs[i] * (ys[i + 1] - ys[i - 1]) for i in range(1, len(coords))) / 2.


def clockwise(coords):
    return signed_area(coords) < 0


def counterclockwise(coords):
    return signed_area(coords) >= 0


def azimuth_distance(x0, y0, x1, y1, longlat=True):
    '''Azimuth and distance between two points'''
    if longlat:
        try:
            az, _, dist = WGS84.inv(x0, y0, x1, y1)
        except NameError:
            raise NotImplementedError("Distance of long/lat coordinates requires pyproj.")

    else:
        az = azimuth(x0, y0, x1, y1, longlat=longlat)
        dist = distance(x0, y0, x1, y1, longlat=longlat)

    return az, dist


def azimuth_distance_points(p0, p1, longlat=True):
    return azimuth_distance(p0[0], p0[1], p1[0], p1[1], longlat)


def det(a, b):
    '''Determinant of a 2x2 matrix'''
    return a[0] * b[1] - a[1] * b[0]


def onsegment(segment, point):
    '''Return True if point lies on segment'''
    xs, ys = tuple(zip(*segment))
    pbr = point[0], point[1], point[0], point[1]

    if not intersectingbounds((min(xs), min(ys), max(xs), max(ys)), pbr):
        return False

    return (point[1] - ys[0]) * (xs[1] - xs[0]) == (ys[1] - ys[0]) * (point[0] - xs[0])


def intersectingbounds(mbr1, mbr2):
    '''Return True if two bounding boxes intersect, else False.'''
    # Check if intersection is impossible
    if mbr1[2] < mbr2[0] or mbr2[2] < mbr1[0] or mbr1[3] < mbr2[1] or mbr2[3] < mbr1[1]:
        return False
    else:
        return True


def coincidentendpoint(linem, linen):
    '''Assuming segments linem and linem, return an overlapping point'''
    if onsegment(linem, linen[0]):
        return linen[0]
    elif onsegment(linem, linen[1]):
        return linen[1]
    elif onsegment(linen, linem[0]):
        return linem[0]
    elif onsegment(linen, linem[1]):
        return linem[1]
    return None


def intersect(linem, linen, detm=None):
    '''
    Check if two line segments intersect. Returns None or the intersection.
    If lines are coincident, returns one of the midpoints.

    Args:
        linem (Sequence): line segment 1, e.g. ((0, 0), (1, 1))
        linen (Sequence): line segment 2, e.g. [[0, 1], [0, 1]]
        det (float): The determinant of linem. Precalculating might be useful if
                   calculating intersection with the same segment many times
    '''
    mx, my = tuple(zip(*linem))
    nx, ny = tuple(zip(*linen))

    mbrm = min(mx), min(my), max(mx), max(my)
    mbrn = min(nx), min(ny), max(nx), max(ny)

    if not intersectingbounds(mbrm, mbrn):
        return None

    xD = mx[0] - mx[1], nx[0] - nx[1]
    yD = my[0] - my[1], ny[0] - ny[1]

    div = det(xD, yD)

    if div == 0:
        try:
            # Check if lines are parallel
            if (xD[0] == xD[1] and yD[0] == yD[1]) or (yD[0] / xD[0] == yD[1] / float(xD[1])):
                return coincidentendpoint(linem, linen)

        except ZeroDivisionError:
            pass

        return None

    else:
        detm = detm or det(*linem)
        detn = det(*linen)
        x = det((detm, detn), xD) / float(div)
        y = det((detm, detn), yD) / float(div)

        if (intersectingbounds(mbrm, (x, y, x, y)) and
                intersectingbounds(mbrn, (x, y, x, y))):
            return x, y
        else:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from unittest import TestCase as PythonTestCase
import unittest.main
from math import pi
from os import path
from fionautil import measure
import fionautil.layer

shp = path.join(path.dirname(__file__), 'fixtures/testing.shp')


class TestMeasure(PythonTestCase):

    def testDistance(self):
        assert measure.distance(1, 0, 0, 0, False) == 1
        self.assertEqual(measure.distance(0, 0, 360, 0, True), 0.0)
        self.assertEqual(measure.distance(0, 0, 6, 0, True), 667916.9447596414)

    def testAzimuth(self):
        self.assertEqual(measure.azimuth(0, 0, 0, 0, clockwise=True, longlat=False), 0)
        self.assertEqual(measure.azimuth(1, 0, 0, 1, clockwise=True, longlat=False), -45)
        self.assertEqual(measure.azimuth(0, 1, 1, 0, clockwise=True, longlat=False), 135)
        self.assertEqual(measure.azimuth(0, 0, 0, 1, clockwise=True, longlat=False), 0)
        self.assertEqual(measure.azimuth(0, 0, 1, 0, clockwise=False, longlat=False), -90)
        self.assertEqual(measure.azimuth(1, 0, 0, 0, clockwise=True, longlat=False), 270)

        self.assertEqual(measure.azimuth(0, 0, 0, 90), -0.0)
        self.assertEqual(measure.azimuth(0, 0, 90, 0), -90.0)
        self.assertEqual(measure.azimuth(0, 0, 90, 0, radians=True), pi / -2)

    def testSignedArea(self):
        feature = fionautil.layer.first(shp)
        self.assertEqual(measure.signed_area(feature['geometry']['coordinates'][0]), -4.428726877457176)

        coords = [(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)]
        assert measure.signed_area(coords) == 1.0

        assert measure.clockwise(coords) == False
        assert measure.counterclockwise(coords) == True

        coords.reverse()

        assert measure.signed_area(coords) == -1.0
        assert measure.clockwise(coords) == True
        assert measure.counterclockwise(coords) == False

        zcoords = [(0, 0, 1), (1, 0, 0), (1, 1, 0), (0, 1, 0), (0, 0, 0)]
        assert measure.signed_area(zcoords) == 1.0

    def testAzimuthDistance(self):
        self.assertEqual(measure.azimuth_distance(0, 0, 90, 0), (90, 10018754.171394622))
        self.assertEqual(measure.azimuth_distance(1, 0, 0, 0, longlat=False), (-270, 1))

    def testDet(self):
        assert measure.det((1, 2), (3, 4)) == -2
        assert measure.det((3, 4), (1, 2)) == 2
        assert measure.det((100, 4), (1, -100)) == -10004

    def testIntersectingbounds(self):
        a = (0, 0, 10, 10)
        b = (0, 0, 9, 9)
        c = (10, 10, 20, 20)

        assert measure.intersectingbounds(a, b) is True
        assert measure.intersectingbounds(a, c) is True
        assert measure.intersectingbounds(b, c) is False

    def testIntersection(self):
        a = ((0, 0), (10, 10))
        b = ((10, 0), (0, 10))
        c = ((0, 5), (10, 5))
        d = ((12, 100), (13, 102))
        e = (0, 0), (5, 5)

        f = ((0, 0), (10, 0))
        g = ((5, 0), (15, 0))

        h = ((0, 0), (0, 10))
        i = ((0, 5), (0, 15))

        j = ((11, 11), (12, 12))

        k = (4, 5), (10, 11)
        m = (10, 10), (0, 0)
        n = (0, 10), (10, 10)

        self.assertEqual(measure.intersect(a, b), (5, 5))
        self.assertEqual(measure.intersect(a, c), (5, 5))

        self.assertIn(measure.intersect(a, e), list(a) + list(e))
        assert measure.intersect(e, a) in list(a) + list(e)

        assert measure.intersect(a, d) is None
        assert measure.intersect(b, d) is None

        assert measure.intersect(f, g) in list(f) + list(g)
        assert measure.intersect(g, f) in list(f) + list(g)

        assert measure.intersect(h, i) in list(h) + list(i)
        assert measure.intersect(i, h) in list(h) + list(i)

        assert measure.intersect(a, j) is None
        assert measure.intersect(k, m) is None
        assert measure.intersect(k, n) == (9, 10)

    def testIntersectionDet(self):
        minx, miny, maxx, maxy = 0, 0, 10, 10
        edges = (
            ((minx, miny), (minx, maxy)),
            ((minx, maxy), (maxx, maxy)),
            ((maxx, maxy), (maxx, miny)),
            ((maxx, miny), (minx, miny))
        )
        dets = [measure.det(*j) for j in edges]
        assert dets == [0, -100, -100, 0]

        a = (4, 5), (10, 11)
        assert measure.intersect(edges[0], a) is None
        assert measure.intersect(edges[1], a) == (9, 10)

        inters = [measure.intersect(e, a, detm=d) for e, d in zip(edges, dets)]
        self.assertListEqual(inters, [None, (9, 10), None, None])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

import unittest
import fionautil.round
try:
    import numpy as np
except ImportError:
    pass


class RoundTestCase(unittest.TestCase):

    def setUp(self):
        self.polygon = {
            "type": "Polygon",
            "coordinates": [
                [
                    [100.11111, 0.0],
                    [101.0, 0.11111],
                    [101.0, 1.11111],
                    [100.11111, 1.0],
                    [100.11111, 0.0]
                ]
            ]
        }

    def testRound(self):
        ring = [(10.00011111, 10.00011111), (10.00011111, 10.00011111)]
        round1 = fionautil.round._round(ring[0], 1)
        round2 = fionautil.round.round_ring(ring, 1)

        try:
            round1 = round1.tolist()
            round2 = round2[0].tolist()
        except AttributeError:
            round2 = round2[0]

        self.assertSequenceEqual(round1, (10.0, 10.0))
        self.assertSequenceEqual(round2, (10.0, 10.0))

    def testRoundPolygon(self):
        g = fionautil.round.geometry(self.polygon, 3)

        try:
            coordinates = g['coordinates'].tolist()
        except AttributeError:
            coordinates = g['coordinates']

        self.assertSequenceEqual(coordinates[0][0], [100.111, 0.0])

    def testRoundMultiPolygon(self):
        mp = {
            "type": "MultiPolygon",
            "coordinates": [self.polygon['coordinates'], self.polygon['coordinates']]
        }
        g = fionautil.round.geometry(mp, 3)

        try:
            coordinates = g['coordinates'][0][0][0].tolist()
        except AttributeError:
            coordinates = g['coordinates'][0][0][0]

        self.assertSequenceEqual(coordinates, [100.111, 0.0])

    def testRoundFeature(self):
        feat = {
            "geometry": self.polygon,
            "properties": {"foo": "bar"}
        }
        f = fionautil.round.feature(feat, 4)

        try:
            coordinates = f['geometry']['coordinates'][0][0].tolist()
        except AttributeError:
            coordinates = f['geometry']['coordinates'][0][0]

        self.assertSequenceEqual(coordinates, [100.1111, 0.0])
        assert f['properties']['foo'] == 'bar'

    def testRoundGenerator(self):
        x = (float(x) + 0.555 for x in range(2))
        b = fionautil.round._round(x, 2)

        try:
            b = b.tolist()
        except AttributeError:
            pass

        self.assertSequenceEqual(b, (0.56, 1.56))

    def testRoundListIfNumpy(self):
        try:
            np
        except NameError:
            return

        result = fionautil.round.geometry(self.polygon, 0)

        self.assertSequenceEqual(result['coordinates'][0][0].tolist(), [100.0, 0.0])

if __name__ == '__main__':
    unittest.main()



# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

import itertools
from functools import reduce
import sys
import fiona
import fiona.transform

try:
    from shapely.geometry import shape as shapelyshape
except ImportError:
    pass
from .geometry import disjointed


def meta(filename):
    '''Return crs and schema for a layer'''
    with fiona.drivers():
        with fiona.open(filename, "r") as layer:
            return layer.meta


def meta_complete(filename):
    '''Return crs and schema for a layer, as well as additional metadata.'''
    with fiona.drivers():
        with fiona.open(filename, "r") as layer:
            m = {
                'bounds': layer.bounds,
                'path': layer.path,
                'name': layer.name,
                'encoding': layer.encoding
            }
            m.update(layer.meta)
            return m


def bounds(filename):
    '''Shortcut for returning bounds of a layer (minx, miny, maxx, maxy)'''
    with fiona.drivers():
        with fiona.open(filename, 'r') as layer:
            return layer.bounds


def first(filename):
    '''Return the first feature of a layer'''
    with fiona.drivers():
        with fiona.open(filename, 'r') as layer:
            return next(layer)


def fiter(filename):
    with fiona.drivers():
        with fiona.open(filename, "r") as layer:
            for feature in layer:
                yield feature


def _fionaiter(iterfunc):
    def _iter(func, filename, *args, **kwargs):
        with fiona.drivers():
            with fiona.open(filename, "r") as layer:
                for feature in iterfunc(func, layer, *args, **kwargs):
                    yield feature
    return _iter


@_fionaiter
def ffilter(func, layer, *args, **kwargs):
    return iter(f for f in layer if func(f, *args, **kwargs))


@_fionaiter
def ffilterfalse(func, layer, *args, **kwargs):
    return iter(f for f in layer if not func(f, *args, **kwargs))


@_fionaiter
def fmap(func, layer, *args, **kwargs):
    '''Yield features in a fiona layer, applying func to each'''
    return iter(func(f, *args, **kwargs) for f in layer)


def freduce(func, filename, initializer=None):
    '''Reduce features of a layer to a single value'''
    with fiona.drivers():
        with fiona.open(filename, "r") as layer:
            return reduce(func, layer, initializer)


def fchain(*filenames):
    '''Reduce features of a layer to a single value'''
    for filename in itertools.chain(filenames):
        with fiona.drivers():
            with fiona.open(filename, "r") as layer:
                for feature in layer:
                    yield feature


def fslice(filename, start, stop=None, step=None):
    if stop is None:
        stop = start
        start = None

    it = iter(range(start or 0, stop or sys.maxsize, step or 1))
    nexti = next(it)

    with fiona.drivers():
        with fiona.open(filename, "r") as layer:
            for i, element in enumerate(layer):
                if i == nexti:
                    yield element
                    nexti = next(it)


def fzip(*filenames):
    with fiona.drivers():
        try:
            handles = [fiona.open(f) for f in filenames]
            for features in zip(*handles):
                yield features

        finally:
            for h in handles:
                h.close()


def shapes(filename, crs=None):
    '''
    Generator that yields a Shapely shape for every feature in a layer.
    '''
    try:
        shapelyshape
    except NameError:
        raise NotImplementedError("length require shapely")

    with fiona.drivers():
        with fiona.open(filename, 'r') as layer:
            if crs is not None:
                def _geom(feature):
                    return fiona.transform.transform_geom(layer.crs, crs, feature['geometry'])
            else:
                def _geom(feature):
                    return feature['geometry']

            for feature in layer:
                yield shapelyshape(_geom(feature))


def length(filename, crs=None):
    '''Get the length of a geodata file in its
    native projection or the given crs mapping'''
    geometries = shapes(filename, crs)
    return sum(x.length for x in geometries)


def perimeter(filename, crs=None):
    '''Get perimeter of all features in a geodata file in its
    native projection or the given crs mapping'''
    geometries = shapes(filename, crs)
    return sum(x.boundary.length for x in geometries)


def find(filename, key, value):
    '''Special case of ffilter: return the first feature where key==value'''
    def test(f):
        return f['properties'][key] == value
    return next(ffilter(test, filename))


def dissolve(sourcefile, sinkfile, key, unsplit=None):
    try:
        shape
    except NameError:
        raise NotImplementedError("dissolve require shapely")

    with fiona.drivers():
        with fiona.open(sourcefile) as source:
            schema = source.schema
            schema['properties'] = {key: source.schema['properties'][key]}

            with fiona.open(sinkfile, 'w', crs=source.crs, schema=schema, driver=source.driver) as sink:

                gotkeys = dict()

                for _, feat in source.items():
                    fkey = feat['properties'][key]
                    fshape = shapelyshape(feat['geometry'])

                    if fkey in gotkeys:
                        gotkeys[fkey][0] = gotkeys[fkey][0].union(fshape)
                    else:
                        gotkeys[fkey] = [fshape]

                for shapelist in gotkeys.values():
                    if unsplit:
                        for s in disjointed(shapelist):
                            sink.write(s)

                    else:
                        sink.write(shapelist[0])



# -*- coding: utf-8 -*-

"""Scale geometries and features."""

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

try:
    import numpy as np
except ImportError:
    pass


def geometry(geom, factor=1):
    g = dict(geom.items())

    if geom['type'] == 'MultiPolygon':
        g['coordinates'] = [scale_rings(rings, factor) for rings in geom['coordinates']]

    elif geom['type'] in ('Polygon', 'MultiLineString'):
        g['coordinates'] = scale_rings(geom['coordinates'], factor)

    elif geom['type'] in ('MultiPoint', 'LineString'):
        g['coordinates'] = scale(geom['coordinates'], factor)

    elif geom['type'] == 'Point':
        g['coordinates'] = scale(geom['coordinates'], factor)

    elif geom['type'] == 'GeometryCollection':
        g['geometries'] = [geometry(i) for i in geom['geometries']]

    else:
        raise NotImplementedError("Unknown geometry type")

    return g


def scale_rings(rings, factor=1):
    return [scale(ring, factor) for ring in rings]


def scale(coordinates, scalar=1):
    '''Scale a list of coordinates by a scalar. Only use with projected coordinates'''
    try:
        try:
            arr = np.array(coordinates, dtype=float)

        except TypeError:
            arr = np.array(list(coordinates), dtype=float)

        return arr * scalar

    except NameError:
        if isinstance(coordinates, tuple):
            return (coordinates[0] * scalar, coordinates[1] * scalar)

        return ((c[0] * scalar, c[1] * scalar) for c in coordinates)


def feature(feat, factor=1):
    return {
        'properties': feat.get('properties'),
        'geometry': geometry(feat['geometry'], factor),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

import collections
from unittest import TestCase as PythonTestCase
import unittest.main
import os.path
import fionautil.layer

shp = os.path.join(os.path.dirname(__file__), 'fixtures/testing.shp')
geojson = os.path.join(os.path.dirname(__file__), 'fixtures/testing.geojson')


class TestLayer(PythonTestCase):

    def test_ffilter(self):
        test = lambda f: f['properties']['id'] == 3
        assert len(list(fionautil.layer.ffilter(test, shp))) == 1
        assert len(list(fionautil.layer.ffilter(test, geojson))) == 1

    def test_ffilterfalse(self):
        test = lambda f: f['properties']['id'] == 3
        assert len(list(fionautil.layer.ffilterfalse(test, shp))) == 3
        assert len(list(fionautil.layer.ffilterfalse(test, shp))) == 3

    def test_fiter(self):
        s = fionautil.layer.fiter(shp)
        d = collections.deque(s, maxlen=0)
        assert len(d) == 0
        assert next(fionautil.layer.fiter(geojson))


    def test_fmap(self):
        def func(f):
            f['properties']['cat'] = 'meow'
            return f

        assert next(fionautil.layer.fmap(func, shp)).get('properties').get('cat') == 'meow'
        assert next(fionautil.layer.fmap(func, geojson)).get('properties').get('cat') == 'meow'

    def test_freduce(self):
        def func(g, f):
            g = g or {'geometry': {'coordinates': []}}
            g['geometry']['coordinates'] = g['geometry']['coordinates'] + f['geometry']['coordinates']
            return g

        start = {
            'geometry': {
                'coordinates': []
            }
        }

        assert fionautil.layer.freduce(func, shp, start)
        assert fionautil.layer.freduce(func, geojson)
        assert len(fionautil.layer.freduce(func, shp, start).get('geometry').get('coordinates')) == 8

    def test_fchain(self):
        chain = fionautil.layer.fchain(shp, geojson)
        assert len(list(chain)) == 8

    def test_fslice(self):
        assert list(fionautil.layer.fslice(shp, 1, 2))
        assert list(fionautil.layer.fslice(geojson, 0, 4, 2))

    def test_fzip(self):
        assert list(fionautil.layer.fzip(shp, geojson))
        assert len(list(fionautil.layer.fzip(shp, geojson))) == 4
        assert type(next(fionautil.layer.fzip(shp, geojson))) == tuple

    def test_find(self):
        two = fionautil.layer.find(shp, 'id', 2)
        one = fionautil.layer.find(geojson, 'id', 1)
        assert one['properties']['id'] == 1
        assert two['properties']['id'] == 2

    def test_layer_args(self):
        def func(f, n):
            f['properties']['n'] = n
            return f

        assert next(fionautil.layer.fmap(func, shp, n='meow')).get('properties').get('n') == 'meow'
        assert next(fionautil.layer.fmap(func, geojson, n='meow')).get('properties').get('n') == 'meow'


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from unittest import TestCase as PythonTestCase
import unittest.main
from fionautil import scale
try:
    import numpy as np
except ImportError:
    pass

class TestCoords(PythonTestCase):

    def setUp(self):
        self.coords = [(1, 1), (-2, 2), (3, 14), (10, -1)]
        self.coordsx2 = [(2, 2), (-4, 4), (6, 28), (20, -2)]

    def test_scale(self):
        result = scale.scale(self.coords, 2)

        try:
            arr = np.array(self.coordsx2, dtype=float).tolist()
            result = result.tolist()
            self.assertEqual(result, arr)

        except NameError:
            self.assertEqual(list(result), self.coordsx2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from setuptools import setup, find_packages

try:
    readme = open('README.rst').read()
except IOError:
    readme = ''

shapely = 'shapely>=1.5.0,<2.0'

with open('fionautil/__init__.py') as i:
    version = next(r for r in i.readlines() if '__version__' in r).split('=')[1].strip('"\' \n')

setup(
    name='fionautil',
    version=version,
    description='helpful utilities for working with geodata with Fiona',
    long_description=readme,
    keywords='GIS',
    author='Neil Freeman',
    author_email='contact@fakeisthenewreal.org',
    url='http://github.com/fitnr/fionautil/',
    license='GNU General Public License v3 (GPLv3)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'fiona>=1.6.3,<2.0',
    ],

    extras_require={
        'shapify': [shapely],
        'shape': [shapely],
        'length': [shapely],
        'dissolve': [shapely],
        'speed': ['numpy>1.9'],
        'azimuth': ['pyproj>=1.9.5,<1.10'],
    },

    test_suite='tests',
)



# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

try:
    import numpy as np
except ImportError:
    pass


def _round(pt, precision):
    try:
        return round(pt[0], precision), round(pt[1], precision)

    except TypeError:
        pt = list(pt)
        return round(pt[0], precision), round(pt[1], precision)


def round_ring(ring, precision):
    return [_round(tuple(pt), precision) for pt in ring]


def geometry(geom, precision):
    g = dict(geom.items())

    if geom['type'] == 'GeometryCollection':
        g['geometries'] = [geometry(x, precision) for x in geom['geometries']]
        return g

    try:
        c = np.array(geom['coordinates'])
        g['coordinates'] = np.round(c, precision)

    except (AttributeError, KeyError, TypeError, NameError):

        if geom['type'] == 'Point':
            g['coordinates'] = _round(geom['coordinates'], precision)

        elif geom['type'] in ('MultiPoint', 'LineString'):
            g['coordinates'] = round_ring(geom['coordinates'], precision)

        elif geom['type'] in ('MultiLineString', 'Polygon'):
            g['coordinates'] = [round_ring(r, precision) for r in geom['coordinates']]

        elif geom['type'] == 'MultiPolygon':
            g['coordinates'] = [[round_ring(r, precision) for r in rings] for rings in geom['coordinates']]

    return g


def feature(feat, precision):
    feat['geometry'] = geometry(feat['geometry'], precision)
    return feat



# -*- coding: utf-8 -*-
# This file is part of fionautil.
# http://github.com/fitnr/fionautil

# Licensed under the GPLv3 license:
# http://http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-6, Neil Freeman <contact@fakeisthenewreal.org>

from . import measure
from .round import geometry as roundgeometry

__all__ = [
    'endpoints',
    'startpoint',
    'endpoint',
    'azimuth',
    'disjointed',
    'explodepoints',
    'explodesegments',
    'exploderings',
    'countpoints',
    'countsegments',
    'roundgeometry'
]


def endpoints(geometry):
    '''Return the first and last coordinates of a LineString or Multilinestring'''
    coords = geometry['coordinates']

    if geometry['type'] == 'MultiLineString':
        return coords[0][0], coords[-1][-1]

    elif geometry['type'] == 'LineString':
        return coords[0], coords[-1]

    raise ValueError("Improper Feature type.")


def startpoint(geometry):
    '''Return the first point in a linear geometry'''
    start, _ = endpoints(geometry)
    return start


def endpoint(geometry):
    '''Return the last point in a linear geometry'''
    _, end = endpoints(geometry)
    return end


def azimuth(geometry, crs=None, radians=None, clockwise=None, longlat=None):
    '''
    Get the azimuth between the start and end points of a polyline geometry.

    Args:
        geometry (dict): A geojson-like geometry object
        crs (dict): A fiona-style mapping
        radians (bool): Output in radians
        clockwise (bool): Use clockwise orientation.
        longlat (bool): geometry's coordinates are in long-lat form.

    Returns:
        (float) The angle (from north) described when one is standing at
                the start of the feature and pointing to the end.
    '''

    if geometry['type'] not in ('LineString', 'MultiLineString'):
        raise ValueError("This only works with PolyLine layers, this is a {type}".format(**geometry))

    crs = '' or crs
    if longlat is None and 'proj' in crs:
        longlat = crs['proj'] == 'longlat'

    first, last = endpoints(geometry)

    points = first + last

    return measure.azimuth(*points, radians=radians, clockwise=clockwise, longlat=longlat)


def disjointed(shapes):
    '''Reduce a list of shapely shapes to those that are disjoint from the others'''
    newshapes = [shapes[0]]
    for shape in shapes[1:]:
        for i, other in enumerate(newshapes):
            if shape.intersects(other):
                newshapes[i] = other.union(shape)
                break

    return newshapes


def explodepoints(geometry):
    '''Generator that returns every coordinate a geometry'''
    if geometry['type'] == 'Point':
        yield geometry['coordinates']

    else:
        for ring in exploderings(geometry):
            for point in ring:
                yield point


def explodesegments(geometry):
    '''Generator that returns every line segment of a geometry'''
    # Sure, could just use explodepoints, but isn't that a
    # bit more memory-intensive, copying all those lists of coords?
    if geometry['type'] in ('MultiLineString', 'Polygon'):
        for ring in geometry['coordinates']:
            for i, point in enumerate(ring[:-1]):
                yield point, ring[i + 1]

    elif geometry['type'] in ('LineString', 'MultiPoint'):
        for i, point in enumerate(geometry['coordinates'][:-1]):
            yield point, geometry['coordinates'][i + 1]

    elif geometry['type'] == 'MultiPolygon':
        for poly in geometry['coordinates']:
            for ring in poly:
                for i, point in enumerate(ring[:-1]):
                    yield point, ring[i + 1]

    elif geometry['type'] == 'GeometryCollection':
        for g in geometry['geometries']:
            for i, j in explodesegments(g):
                yield i, j

    else:
        raise ValueError("Unknown or invalid geometry type: {type}".format(**geometry))


def exploderings(geometry):
    '''Generator that returns every ring of a geometry (a ring is a list of points).'''
    if geometry['type'] in ('MultiLineString', 'Polygon'):
        for ring in geometry['coordinates']:
            yield ring

    elif geometry['type'] in ('LineString', 'MultiPoint'):
        yield geometry['coordinates']

    elif geometry['type'] == 'MultiPolygon':
        for poly in geometry['coordinates']:
            for ring in poly:
                yield ring

    elif geometry['type'] == 'GeometryCollection':
        for g in geometry['geometries']:
            for ring in exploderings(g):
                yield ring

    else:
        raise ValueError("Unknown or invalid geometry type: {type}".format(**geometry))


def countpoints(geometry):
    '''Returns the number of points in a geometry'''
    if geometry['type'] == 'Point':
        return 1

    else:
        return sum(len(ring) for ring in exploderings(geometry))


def countsegments(geometry):
    '''Not guaranteed for (multi)point layers'''
    if geometry['type'] == 'Point':
        return 0

    else:
        return sum(len(ring) - 1 for ring in exploderings(geometry))
